# -*- coding: utf-8 -*-
"""
*A similarity preserving perceptual hash for images.*

The ISCC Content-Code Image is created by calculating a discrete cosine transform on
normalized image-pixels and comparing the values from the upper left area of the
dct-matrix against their median values to set the hash-bits.
"""
from statistics import median
from typing import Sequence, Tuple
from PIL import Image
from more_itertools import chunked
from iscc_core import codec
from iscc_core.codec import Stream
from iscc_core.models import ContentCodeImage
from iscc_core.options import opts
from iscc_core.dct import dct


def gen_image_code(img, bits=opts.image_bits):
    # type: (Stream, int) -> ContentCodeImage
    """Create an ISCC Content-Code Image with the latest standard algorithm.

    :param Stream img: Image data stream.
    :param int bits: Bit-length of ISCC Content-Code Image (default 64).
    :return: ISCC Content-Code Image.
    :rtype: ContentCodeImage
    """
    return gen_image_code_v0(img, bits)


def gen_image_code_v0(img, bits=opts.image_bits):
    # type: (Stream, int) -> ContentCodeImage
    """Create an ISCC Content-Code Image with algorithm v0.

    :param Stream img: Image data stream.
    :param int bits: Bit-length of ISCC Content-Code Image (default 64).
    :return: ISCC Content-Code Image.
    :rtype: ContentCodeImage
    """
    pixels, width, height = normalize_image(img)
    digest = soft_hash_image_v0(pixels, bits=bits)
    image_code = codec.encode_component(
        mtype=codec.MT.CONTENT,
        stype=codec.ST_CC.IMAGE,
        version=codec.VS.V0,
        length=bits,
        digest=digest,
    )
    return ContentCodeImage(code=image_code, width=width, height=height)


def normalize_image(img):
    # type: (Stream) -> Tuple[Sequence[int], int, int]
    """Normalize image for hash calculation.

    - Add white background to images with alpha transparency
    - Convert to grayscale
    - Resize (bicubic) to 32x32 pixels and flatten

    :param Stream img: Image data stream.
    :return: A tuple of (pixels_1024, original_width, original_height
    :rtype: Tuple[Sequence[int], int, int]
    """

    im = Image.open(img)
    width, height = im.size

    # Add white background to images that have alpha transparency
    if im.mode in ("RGBA", "LA") or (im.mode == "P" and "transparency" in im.info):
        alpha = im.convert("RGBA").split()[-1]
        bg = Image.new("RGBA", im.size, (255, 255, 255, 255))
        bg.paste(im, mask=alpha)
        im = bg

    # Convert to grayscale
    im = im.convert("L")

    # Resize to 32x32
    im = im.resize((32, 32), Image.BICUBIC)

    # A flattened sequence of grayscale pixel values (1024 pixels)
    pixels = im.getdata()

    return pixels, width, height


def soft_hash_image_v0(pixels, bits=opts.image_bits):
    # type: (Sequence[int], int) -> bytes
    """Calculate image hash from normalized grayscale pixel sequence of length 1024.

    :param Sequence[int] pixels:
    :param int bits: Bit-length of image hash (default 64).
    :return: Similarity preserving Image-Hash digest.
    :rtype: bytes
    """

    assert bits <= 256

    # DCT per row
    dct_row_lists = []
    for pixel_list in chunked(pixels, 32):
        dct_row_lists.append(dct(pixel_list))

    # DCT per col
    dct_row_lists_t = list(map(list, zip(*dct_row_lists)))
    dct_col_lists_t = []
    for dct_list in dct_row_lists_t:
        dct_col_lists_t.append(dct(dct_list))

    dct_matrix = list(map(list, zip(*dct_col_lists_t)))

    def flatten(m, x, y):
        """Extract and flatten an 8 x 8 slice from a 2d matrix starting at col/row."""
        return [v for sublist in m[y : y + 8] for v in sublist[x : x + 8]]

    bitstring = ""
    slices = ((0, 0), (1, 0), (0, 1), (1, 1))

    for xy in slices:
        # Extract 8 x 8 slice
        flat_list = flatten(dct_matrix, *xy)

        # Calculate median
        med = median(flat_list)

        # Append 64-bit digest by comparing to median
        for value in flat_list:
            if value > med:
                bitstring += "1"
            else:
                bitstring += "0"
        bl = len(bitstring)
        if bl >= bits:
            hash_digest = int(bitstring, 2).to_bytes(bl // 8, "big", signed=False)
            return hash_digest



# -*- coding: utf-8 -*-
"""
# ISCC Content-Code Video

The **Content-Code Video** is generated from MPEG-7 Video Frame Signatures.
Frame Signatures can be extracted with ffmpeg (see: https://www.ffmpeg.org/) using the
following command line parameters:

`$ ffmpeg -i video.mpg -vf fps=fps=5,signature=format=xml:filename=sig.xml -f null -`

The relevant frame signatures can be parsed from the following elements in sig.xml:

`<FrameSignature>0  0  0  1  0  0  1  0  1  1  0  0  1  1 ...</FrameSignature>`

!!! note
    it is also possible to extract the signatures in a more compact binary format
    but it requires a custom binary parser to decode the frame signaturs.
"""
from typing import Sequence, Tuple
from iscc_core.wtahash import wtahash
from iscc_core import codec
from iscc_core.options import opts
from iscc_core.models import ContentCodeVideo


FrameSig = Tuple[int]


def gen_video_code(frame_sigs, bits=opts.video_bits):
    # type: (Sequence[FrameSig], int) -> ContentCodeVideo
    """Create an ISCC Content-Code Video with the latest standard algorithm.

    :param FrameSig frame_sigs: Sequence of MP7 frame signatures
    :param int bits: Bit-length resulting Instance-Code (multiple of 64)
    :return: VideoCode object with code property set
    :rtype: ContentCodeVideo
    """
    return gen_video_code_v0(frame_sigs, bits)


def gen_video_code_v0(frame_sigs, bits=opts.video_bits):
    # type: (Sequence[FrameSig], int) -> ContentCodeVideo
    """Create an ISCC Content-Code Video with algorithm v0.

    :param FrameSig frame_sigs: Sequence of MP7 frame signatures
    :param int bits: Bit-length resulting Instance-Code (multiple of 64)
    :return: VideoCode object with code property set
    :rtype: ContentCodeVideo
    """
    digest = soft_hash_video_v0(frame_sigs)
    video_code = codec.encode_component(
        mtype=codec.MT.CONTENT,
        stype=codec.ST_CC.VIDEO,
        version=codec.VS.V0,
        length=bits,
        digest=digest,
    )
    video_code_obj = ContentCodeVideo(code=video_code)
    return video_code_obj


def soft_hash_video_v0(frame_sigs):
    # type: (Sequence[Tuple[int]]) -> bytes
    """Compute 256-bit video hash v0 from MP7 frame signatures."""
    sigs = set(frame_sigs)
    vecsum = [sum(col) for col in zip(*sigs)]
    video_hash_digest = wtahash(vecsum)
    return video_hash_digest



# -*- coding: utf-8 -*-
"""
*A similarity perserving hash for binary data (soft hash).*
"""
from typing import List, Optional, Tuple
from more_itertools import chunked
from iscc_core.cdc import data_chunks
from iscc_core.minhash import minhash_256, minhash_64
from iscc_core import codec
from iscc_core.codec import Data, Stream, encode_base64
from iscc_core.options import opts
from iscc_core.models import DataCode
import xxhash


def gen_data_code(stream, bits=opts.data_bits, granular=opts.data_granular):
    # type: (Stream, int) -> DataCode
    """
    Create a similarity preserving ISCC Data-Code with the latest standard algorithm.

    :param Stream stream: Input data stream.
    :param int bits: Bit-length of ISCC Data-Code (default 64).
    :param bool granular: Calculate additional granular data-features
    :return: ISCC Data-Code with properties: code, features, sizes
    :rtype: DataCode
    """
    return gen_data_code_v0(stream, bits, granular)


def gen_data_code_v0(stream, bits=opts.data_bits, granular=opts.data_granular):
    # type: (Stream, int) -> DataCode
    """
    Create an ISCC Data-Code with algorithm v0.

    :param Stream stream: Input data stream.
    :param int bits: Bit-length of ISCC Data-Code (default 64).
    :param bool granular: Calculate additional granular data-features
    :return: ISCC DataCode with properties: code, features, sizes
    :rtype: DataCode
    """
    features, sizes = None, None
    result = soft_hash_data_v0(stream, granular=granular)
    if granular:
        digest, features, sizes = result
    else:
        digest = result

    data_code = codec.encode_component(
        mtype=codec.MT.DATA,
        stype=codec.ST.NONE,
        version=codec.VS.V0,
        length=bits,
        digest=digest,
    )
    data_code_obj = DataCode(
        code=data_code,
    )

    if granular:
        data_code_obj.features = features
        data_code_obj.sizes = sizes

    return data_code_obj


def soft_hash_data_v0(stream, granular=opts.data_granular):
    # type: (Stream) -> Union[bytes, Tuple[bytes, List[str], List[int]]]
    """
    Create a similarity preserving Data-Hash digest

    :param stream: Input data stream.
    :param bool granular: Calculate additional granular data-features
    :return: 256-bit data soft-hash (if granular: a tuple of digest, features, sizes)
    :rtype: Union[bytes, Tuple[bytes, List[str], List[int]]]
    """
    hasher = DataHasherV0()
    data = stream.read(opts.cdc_read_size)

    while data:
        hasher.push(data)
        data = stream.read(opts.cdc_read_size)

    if granular is False:
        return hasher.digest()
    elif granular is True:
        return hasher.digest(), hasher.features(), hasher.sizes()
    else:
        raise ValueError("parameter granular must be True or False")


class DataHasherV0:
    """Incremental Data-Hash generator."""

    def __init__(self, data=None):
        # type: (Optional[Data]) -> None
        self.chunk_features = []
        self.chunk_sizes = []
        self.tail = None
        data = data or b""
        self.push(data)

    def push(self, data):
        # type: (Data) -> None
        """Push data to the Data-Hash generator."""
        if self.tail:
            data = self.tail + data

        for chunk in data_chunks(
            data, utf32=False, avg_chunk_size=opts.data_avg_chunk_size
        ):
            self.chunk_sizes.append(len(chunk))
            self.chunk_features.append(xxhash.xxh32_intdigest(chunk))
            self.tail = chunk  # Last chunk may not be final

        self.chunk_features = self.chunk_features[:-1]
        self.chunk_sizes = self.chunk_sizes[:-1]

    def digest(self):
        # type: () -> bytes
        """Calculate 256-bit minhash digest from feature hashes."""
        self._finalize()
        return minhash_256(self.chunk_features)

    def features(self):
        # type: () -> List[str]
        """Calculate and encode granular feature hashes.

        :return: Base64 encoded granular features (64-bit soft hashes)
        :rtype: List[str]
        """
        self._finalize()
        encoded_features = [
            encode_base64(minhash_64(cf))
            for cf in chunked(self.chunk_features, opts.data_granular_factor)
        ]
        return encoded_features

    def sizes(self):
        # type: () -> List[int]
        """Calculate sizes of granular feature chunks

        :return: List of sizes of granular features in number of bytes
        :rtype: List[int]
        """
        self._finalize()
        sizes = [sum(fh) for fh in chunked(self.chunk_sizes, opts.data_granular_factor)]
        return sizes

    def _finalize(self):
        if self.tail is not None:
            self.chunk_features.append(xxhash.xxh32_intdigest(self.tail))
            self.chunk_sizes.append(len(self.tail))
            self.tail = None


DataHasher = DataHasherV0



# -*- coding: utf-8 -*-
import blake3
from typing import Optional, Tuple
from iscc_core.options import opts
from iscc_core import codec
from iscc_core.codec import Data, Stream
from iscc_core.models import InstanceCode


def gen_instance_code(stream, bits=opts.instance_bits):
    # type: (Stream, int) -> InstanceCode
    """
    Create an ISCC Instance-Code with the latest standard algorithm.

    :param Stream stream: Binary data stream for Instance-Code generation
    :param int bits: Bit-length resulting Instance-Code (multiple of 64)
    :return: InstanceCode with properties: code, datahash, filesize
    :rtype: InstanceCode
    """
    return gen_instance_code_v0(stream, bits)


def gen_instance_code_v0(stream, bits=opts.instance_bits):
    # type: (Stream, int) -> InstanceCode
    """
    Create an ISCC Instance-Code with algorithm v0.

    :param Stream stream: Binary data stream for Instance-Code generation
    :param int bits: Bit-length of resulting Instance-Code (multiple of 64)
    :return: InstanceCode with properties: code, datahash, filesize
    :rtype: InstanceCode
    """
    digest, filesize = hash_instance_v0(stream)
    instance_code = codec.encode_component(
        mtype=codec.MT.INSTANCE,
        stype=codec.ST.NONE,
        version=codec.VS.V0,
        length=bits,
        digest=digest,
    )

    instance_code_obj = InstanceCode(
        code=instance_code,
        datahash=digest.hex(),
        filesize=filesize,
    )
    return instance_code_obj


def hash_instance_v0(stream):
    # type: (Stream) -> Tuple[bytes, int]
    """
    Create 256-bit hash digest for the Instance-Code body

    :param Stream stream: Binary data stream for hash generation.
    :return: Tuple of 256-bit Instance-Hash digest and filesize in bytes
    :rtype: Tuple[bytes, int]
    """
    hasher = InstanceHasherV0()
    data = stream.read(opts.instance_read_size)
    while data:
        hasher.push(data)
        data = stream.read(opts.instance_read_size)
    return hasher.digest(), hasher.filesize


class InstanceHasherV0:
    """
    Incremental Instance-Hash generator.
    """

    def __init__(self, data=None):
        # type: (Optional[Data]) -> None
        self.hasher = blake3.blake3()
        self.filesize = 0
        data = data or b""
        self.push(data)

    def push(self, data):
        # type: (Data) -> None
        """
        Push data to the Instance-Hash generator.

        :param Data data: Data to be hashed
        """
        self.filesize += len(data)
        self.hasher.update(data)

    def digest(self):
        # type: () -> bytes
        """
        Return Instance-Hash

        :return: Instance-Hash digest
        :rtype: bytes
        """
        return self.hasher.digest()


InstanceHasher = InstanceHasherV0



# -*- coding: utf-8 -*-
"""
The Meta-Code is the first component of a canonical ISCC. It is calculated as a
similarity preserving hash from the metadata of a digital asset. The purpose of the
Meta-Code is the discovery of digital assets with similar metadata or spelling mistakes.

The metadata supplied to the algorithm is called *seed metadata*.

*Seed metadata* is composed of a `title` and an optional generic `extra`-field that
contains descriptive, industry-sector or use-case specific metadata in textual or
binary form (e.g. file headers). We do not prescribe a particular schema.
"""
from more_itertools import interleave, sliced
from iscc_core.code_content_text import normalize_text
from iscc_core.codec import MT, ST, VS, encode_base64, encode_component
from iscc_core.models import MetaCode
from iscc_core.utils import sliding_window
from iscc_core.simhash import similarity_hash
from iscc_core.options import opts
from blake3 import blake3
from typing import Union


def gen_meta_code(title, extra=None, bits=opts.meta_bits):
    # type: (str, Union[str,bytes,None], int) -> MetaCode
    """Create an ISCC Meta-Code using the latest standard algorithm.

    Applications that generate ISCCs should prioritize explicitly passed `title`
    information. If not available they should try to extract a title form the digital
    asset itself. If extraction fails, the application should resort to a normalized
    filename before falling back to an empty string.

    Optional additional metadata may be supplied via the `extra`-field.
    The input can be:

    - A textual description of the identified work for disambiguation purposes
    - Structured (JSON) metadata conforming to an industry specific metadata schema
    - Raw bitstream file headers automatically extracted binary file headers
    - A pre-existing industry-specific identifier string

    !!! note
        It is recommended to use the minimal metadata required to disambiguate the work
        manifested by the digital asset.

    :param str title: Title of the work manifested by the digital asset
    :param Union[str,bytes,None] extra: Optional metadata for disambiguation
    :param int bits: Bit-length of resulting Meta-Code (multiple of 64)
    :return: ISCC Meta-Code
    :rtype: MetaCode
    """
    return gen_meta_code_v0(title, extra=extra, bits=bits)


def gen_meta_code_v0(title, extra=None, bits=opts.meta_bits):
    # type: (str, Union[str,bytes,None], int) -> MetaCode
    """Create an ISCC Meta-Code with the algorithm version 0.

    :param str title: Title of the work manifested by the digital asset
    :param Union[str,bytes,None] extra: Optional metadata for disambiguation
    :param int bits: Bit-length of resulting Meta-Code (multiple of 64)
    :return: ISCC Meta-Code
    :rtype: MetaCode
    """

    # 1. Normalize title
    title = "" if title is None else title
    title = normalize_text(title)
    title = trim_text(title, opts.meta_trim_title)

    payload = title.encode("utf-8")

    # 2. Normalize extra
    if extra in (None, ""):
        extra = None
    elif isinstance(extra, str):
        extra = normalize_text(extra)
        extra = trim_text(extra, opts.meta_trim_extra)
        payload += extra.encode("utf-8")
    elif isinstance(extra, bytes):
        extra = extra[: opts.meta_trim_extra]
        payload += extra
    else:
        raise ValueError("parameter `extra` must be of type str or bytes!")

    digest = soft_hash_meta_v0(title, extra)
    meta_code = encode_component(
        mtype=MT.META,
        stype=ST.NONE,
        version=VS.V0,
        length=bits,
        digest=digest,
    )

    metahash = blake3(payload).hexdigest()

    if isinstance(extra, bytes):
        extra = encode_base64(extra)
        binary = True
    else:
        binary = False

    mc_obj = MetaCode(
        code=meta_code, title=title, extra=extra, binary=binary, metahash=metahash
    )
    return mc_obj


def soft_hash_meta_v0(title, extra=None):
    # type: (str, Union[str,bytes,None]) -> bytes
    """Calculate simmilarity preserving 256-bit hash digest from asset metadata.

    Textual input should be stripped of markup, normalized and trimmed before hashing.
    Json metadata should be normalized with
    [JCS](https://tools.ietf.org/id/draft-rundgren-json-canonicalization-scheme-00.html)

    !!! note
        The processing algorithm depends on the type of the `extra` input.
        If the `extra` field is supplied and non-empty, we create separate hashes for
        `title` and `extra` and interleave them in 32-bit chunks:

        - If the input is `None` or an empty `str`/`bytes` object the Meta-Hash will
        be generated from the `title`-field only.

        - If the `extra`-input is a non-empty **text** string (str) the string is
        lower-cased and the processing unit is an utf-8 endoded character
        (possibly multibyte). The resulting hash is interleaved with the `title`-hash.

        - If the `extra`-input is a non-empty **bytes** object the processing is done
        bytewise and the resulting hash is interleaved with the `title`-hash.

    :param str title: Title of the work manifested in the digital asset
    :param Union[str,bytes,None] extra: Additional metadata for disambiguation
    :return: 256-bit simhash digest for Meta-Code
    :rtype: bytes
    """
    title = title.lower()
    title_n_grams = sliding_window(title, width=opts.meta_ngram_size_title)
    title_hash_digests = [blake3(s.encode("utf-8")).digest() for s in title_n_grams]
    simhash_digest = similarity_hash(title_hash_digests)

    if extra in (None, "", b""):
        return simhash_digest
    else:
        # Augment with interleaved hash for extra metadata
        if isinstance(extra, bytes):
            # Raw bytes are handled per byte
            extra_n_grams = sliding_window(
                extra, width=opts.meta_ngram_size_extra_binary
            )
            extra_hash_digests = [blake3(ngram).digest() for ngram in extra_n_grams]
        elif isinstance(extra, str):
            # Text is lower cased and handled per character (multibyte)
            extra = extra.lower()
            extra_n_grams = sliding_window(extra, width=opts.meta_ngram_size_extra_text)
            extra_hash_digests = [
                blake3(s.encode("utf-8")).digest() for s in extra_n_grams
            ]
        else:
            raise ValueError("parameter `extra` must be of type str or bytes!")

        extra_simhash_digest = similarity_hash(extra_hash_digests)

        # Interleave first half of title and extra simhashes in 32-bit chunks
        chunks_simhash_digest = sliced(simhash_digest[:16], 4)
        chunks_extra_simhash_digest = sliced(extra_simhash_digest[:16], 4)
        interleaved = interleave(chunks_simhash_digest, chunks_extra_simhash_digest)
        simhash_digest = bytearray()
        for chunk in interleaved:
            simhash_digest += chunk

        simhash_digest = bytes(simhash_digest)

        return simhash_digest


def trim_text(text, nbytes):
    # type: (str, int) -> str
    """Trim text such that its utf-8 encoded size does not exceed `nbytes`."""
    return text.encode("utf-8")[:nbytes].decode("utf-8", "ignore").strip()



# -*- coding: utf-8 -*-
"""Schema of objects returned by ISCC processing algorithms"""
from typing import List, Optional, Union
from pydantic import BaseModel, Field
from iscc_core.codec import Code


MultiStr = Union[str, List[str]]


class BaseCode(BaseModel):
    @property
    def code_obj(self):
        return Code(self.code)


class MetaCode(BaseCode):
    """Meta-Code standardized metadata model."""

    code: str = Field(..., description="Meta-Code in standard representation.")
    title: Optional[str] = Field(description="Title used for Meta-Code creation.")
    extra: Optional[str] = Field(description="Extra metadata used for Meta-Code.")
    binary: bool = Field(description="Extra metadata was supplied in binary format.")
    metahash: str = Field(description="Blake3 cryptographic hash of metadata.")


class ContentCodeText(BaseCode):
    """Content-Code-Text standardized metadata model."""

    code: str = Field(..., description="Content-Code-Text in standard representation.")
    title: Optional[str] = Field(description="Title as extracted from text document.")
    characters: Optional[int] = Field(
        description="Number of text characters (after normalize_text)."
    )
    language: Optional[str] = Field(description="Main language of content (BCP-47).")


class ContentCodeImage(BaseCode):
    """Content-Code-Image standardized metadata model."""

    code: str = Field(..., description="Content-Code-Image in standard representation.")
    title: Optional[str] = Field(description="Title as extracted from image file.")
    width: Optional[int] = Field(description="Width of image in number of pixels.")
    height: Optional[int] = Field(description="Height of image in number of pixels.")


class ContentCodeAudio(BaseCode):
    """Content-Code-Audio standardized metadata model."""

    code: str = Field(..., description="Content-Code-Audio in standard representation.")
    title: Optional[str] = Field(description="Title as extracted from audio asset.")
    duration: Optional[float] = Field(description="Duration of audio im seconds.")


class ContentCodeVideo(BaseCode):
    """Content-Code-Video standardized metadata model."""

    code: str = Field(..., description="Content-Code-Video in standard representation.")
    title: Optional[str] = Field(description="Title as extracted from video asset.")
    duration: Optional[float] = Field(description="Duration of video im seconds.")
    fps: Optional[float] = Field(description="Frames per second.")
    width: Optional[int] = Field(description="Width of video in number of pixels.")
    height: Optional[int] = Field(description="Height of video in number of pixels.")
    language: Optional[MultiStr] = Field(description="Main language of video (BCP 47).")


class DataCode(BaseCode):
    """Data-Code standardized metadata model."""

    code: str = Field(..., description="Data-Code in standard representation.")
    features: Optional[List[str]] = Field(description="List of per datachunk hashes")
    sizes: Optional[List[int]] = Field(description="Sizes of datachunks")


class InstanceCode(BaseCode):
    """Instance-Code standardized metadata model."""

    code: str = Field(..., description="Instance-Code in standard representation.")
    datahash: str = Field(description="Multihash of digital asset (Blake3 by default.")
    filesize: int = Field(description="File size in bytes.")



# -*- coding: utf-8 -*-
"""
*A similarity preserving hash for audio content (soft hash).*

The Content-Code Audio is generated from a [Chromaprint](https://acoustid.org/chromaprint)
fingerprint provided as a vector of 32-bit signed integers.
Chromaprints are extracted with [fpcalc 1.5.0](https://acoustid.org/chromaprint)
using the following command line parameters:

`$ fpcalc -raw -json -signed -length 0 myaudiofile.mp3`
"""
from typing import Iterable
from more_itertools import divide

from iscc_core.models import ContentCodeAudio
from iscc_core.simhash import similarity_hash
from iscc_core import codec
from iscc_core.options import opts


def gen_audio_code(cv, bits=opts.audio_bits):
    # type: (Iterable[int], int) -> ContentCodeAudio
    """Create an ISCC Content-Code Audio with the latest standard algorithm.

    :param Iterable[int] cv: Chromaprint vector
    :param int bits: Bit-length resulting Content-Code Audio (multiple of 64)
    :return: ContentCodeAudio object
    :rtype: ContentCodeAudio
    """
    return gen_audio_code_v0(cv, bits)


def gen_audio_code_v0(cv, bits=opts.audio_bits):
    # type: (Iterable[int], int) -> ContentCodeAudio
    """Create an ISCC Content-Code Audio with algorithm v0.

    :param Iterable[int] cv: Chromaprint vector
    :param int bits: Bit-length resulting Content-Code Audio (multiple of 64)
    :return: ContentCodeAudio object
    :rtype: ContentCodeAudio
    """
    digest = soft_hash_audio_v0(cv)
    audio_code = codec.encode_component(
        mtype=codec.MT.CONTENT,
        stype=codec.ST_CC.AUDIO,
        version=codec.VS.V0,
        length=bits,
        digest=digest,
    )
    return ContentCodeAudio(code=audio_code)


def soft_hash_audio_v0(cv):
    # type: (Iterable[int]) -> bytes
    """Create 256-bit audio similarity hash from a chromaprint vector.

    :param Iterable[int] cv: Chromaprint vector
    :return: 256-bit Audio-Hash digest
    :rtype: bytes
    """

    # Convert chrompaprint vector into list of 4 byte digests
    digests = [int_feature.to_bytes(4, "big", signed=True) for int_feature in cv]

    # Return identity hash if we have 0 digests
    if not digests:
        return b"\x00" * 32

    # Calculate simhash of digests as first 32-bit chunk of the hash
    parts = [similarity_hash(digests)]

    # Calculate separate 32-bit simhashes for each quarter of features (original order)
    for bucket in divide(4, digests):
        features = list(bucket)
        if features:
            parts.append(similarity_hash(features))
        else:
            parts.append(b"\x00\x00\x00\x00")

    # Calculate separate simhashes for each third of features (ordered by int value)
    cvs = sorted(cv)
    digests = [int_feature.to_bytes(4, "big", signed=True) for int_feature in cvs]
    for bucket in divide(3, digests):
        features = list(bucket)
        if features:
            parts.append(similarity_hash(features))
        else:
            parts.append(b"\x00\x00\x00\x00")
    return b"".join(parts)



# -*- coding: utf-8 -*-
"""
The `iscc_core.codec` module provides encoding, decoding and transcoding related functions.

#ISCC Component Structure:

**Header:** `<type> <subtype> <version> <length>` each coded as a variable-length 4-bit
sequence.

**Body:** `<hash-digest>` with number of bits as indicated by `<length>`
"""
import enum
import math
import mmap
from io import BufferedReader, BytesIO
from operator import attrgetter
from os import urandom
from random import choice
from typing import BinaryIO, Iterable, List, Tuple, Union
from bitarray import bitarray, frozenbitarray
from bitarray._util import count_xor
from bitarray.util import ba2hex, int2ba, ba2int
from base64 import b32encode, b32decode

try:
    from pybase64 import urlsafe_b64encode, urlsafe_b64decode
except ImportError:
    from base64 import urlsafe_b64encode, urlsafe_b64decode


########################################################################################
# Type definitions                                                                     #
########################################################################################


class MT(enum.IntEnum):
    """ISCC MainTypes"""

    META = 0
    SEMANTIC = 1
    CONTENT = 2
    DATA = 3
    INSTANCE = 4
    ISCC = 5


class ST(enum.IntEnum):
    """Generic SubTypes"""

    NONE = 0


class ST_CC(enum.IntEnum):
    """SubTypes for Content-Codes"""

    TEXT = 0
    IMAGE = 1
    AUDIO = 2
    VIDEO = 3


class VS(enum.IntEnum):
    """ISCC code Versions"""

    V0 = 0


class LN(enum.IntEnum):
    """ISCC length in bits"""

    L32 = 32
    L64 = 64
    L128 = 128
    L256 = 256


Data = Union[bytes, bytearray, memoryview]
Stream = Union[BinaryIO, mmap.mmap, BytesIO, BufferedReader]
IsccTuple = Tuple[
    Union[int, MT], Union[int, ST, ST_CC], Union[int, VS], Union[int, LN], bytes
]
AnyISCC = Union[str, IsccTuple, bytes, "Code"]


########################################################################################
# Core codec functions                                                                 #
########################################################################################


def write_header(mtype, stype, version=0, length=64):
    # type: (int, int, int, int) -> bytes
    """
    Encodes header values with nibble-sized (4-bit) variable-length encoding.
    The result is minimum 2 and maximum 8 bytes long. If the final count of nibbles
    is uneven it is padded with 4-bit `0000` at the end.

    :param int mtype: Main-type of component.
    :param int stype: Sub-type of component.
    :param int version: Version of component algorithm.
    :param int length: Length of component in number of bits (multiple of 32)
    :return: Byte encoded ISCC header.
    :rtype: bytes

    """
    assert length >= 32 and not length % 32, "Length must be a multiple of 32"
    length = (length // 32) - 1
    header = bitarray()
    for n in (mtype, stype, version, length):
        header += _write_varnibble(n)
    # Append zero-padding if required (right side, least significant bits).
    header.fill()
    return header.tobytes()


def read_header(data):
    # type: (bytes) -> Tuple
    """
    Decodes varnibble encoded header and returns it together with remaining bytes.
    :param bytes data: ISCC bytes digest
    :return: (type, subtype, version, length, remaining bytes)
    """
    result = []
    ba = bitarray()
    ba.frombytes(data)
    data = ba
    for x in range(3):
        value, data = _read_varnibble(data)
        result.append(value)

    # interpret length value
    length, data = _read_varnibble(data)

    bit_length = (length + 1) * 32
    result.append(bit_length)

    # Strip 4-bit padding if required
    if len(data) % 8 and data[:4] == bitarray("0000"):
        data = data[4:]

    result.append(data.tobytes())

    return tuple(result)


def encode_base32(data):
    # type: (bytes) -> str
    """
    Standard RFC4648 base32 encoding without padding.
    """
    return b32encode(data).decode("ascii").rstrip("=")


def decode_base32(code):
    # type: (str) -> bytes
    """
    Standard RFC4648 base32 decoding without padding and with casefolding.
    """
    # python stdlib does not support base32 without padding, so we have to re-pad.
    cl = len(code)
    pad_length = math.ceil(cl / 8) * 8 - cl

    return bytes(b32decode(code + "=" * pad_length, casefold=True))


def encode_base64(data: bytes) -> str:
    """
    Standard RFC4648 base64url encoding without padding.
    """
    code = urlsafe_b64encode(data).decode("ascii")
    return code.rstrip("=")


def decode_base64(code: str) -> bytes:
    """
    Standard RFC4648 base64url decoding without padding.
    """
    padding = 4 - (len(code) % 4)
    string = code + ("=" * padding)
    return urlsafe_b64decode(string)


def _write_varnibble(n):
    # type: (int) -> bitarray
    """
    Writes integer to variable length sequence of 4-bit chunks.
    Variable-length encoding scheme:
    ------------------------------------------------------
    | prefix bits | nibbles | data bits | unsigned range |
    | ----------- | ------- | --------- | -------------- |
    | 0           | 1       | 3         | 0 - 7          |
    | 10          | 2       | 6         | 8 - 71         |
    | 110         | 3       | 9         | 72 - 583       |
    | 1110        | 4       | 12        | 584 - 4679     |
    """
    if 0 <= n < 8:
        return int2ba(n, length=4)
    elif 8 <= n < 72:
        return bitarray("10") + int2ba(n - 8, length=6)
    elif 72 <= n < 584:
        return bitarray("110") + int2ba(n - 72, length=9)
    elif 584 <= n < 4680:
        return bitarray("1110") + int2ba(n - 584, length=12)
    else:
        raise ValueError("Value must be between 0 and 4679")


def _read_varnibble(b):
    # type: (bitarray) -> Tuple[int, bitarray]
    """Reads first varnibble, returns its integer value and remaining bits."""

    bits = len(b)

    if bits >= 4 and b[0] == 0:
        return ba2int(b[:4]), b[4:]
    if bits >= 8 and b[1] == 0:
        return ba2int(b[2:8]) + 8, b[8:]
    if bits >= 12 and b[2] == 0:
        return ba2int(b[3:12]) + 72, b[12:]
    if bits >= 16 and b[3] == 0:
        return ba2int(b[4:16]) + 584, b[16:]

    raise ValueError("Invalid bitarray")


########################################################################################
# Convenience functions and classes                                                    #
########################################################################################


def encode_component(mtype, stype, version, length, digest):
    # type: (int, int, int, int, bytes) -> str
    """Encode a ISCC component inlcuding header and body with standard base32 encoding.

    !!! note
        If `digest` has more bits than specified by `length` it wil be truncated.

    :param int mtype: Main-type of component.
    :param int stype: Sub-type of component.
    :param int version: Version of component algorithm.
    :param int length: Length of component in number of bits (multiple of 32)
    :param bytes digest: The hash digest of the component.
    :return: Base32 encoded component code.
    :rtype: str
    """
    nbytes = length // 8
    header = write_header(mtype, stype, version, length)
    body = digest[:nbytes]
    component_code = encode_base32(header + body)
    return component_code


def clean(iscc):
    # type: (str) -> str
    """Cleanup ISCC String.

    Removes leading scheme and dashes.
    """
    return iscc.split(":")[-1].strip().replace("-", "")


class Code:
    """Convenience class to handle different representations of an ISCC."""

    def __init__(self, code):
        # type: (AnyISCC) -> None
        """Initialize a Code object from any kind of representation of an ISCC.

        :param AnyISCC code: Any valid representation of an ISCC
        """
        self._head = None
        self._body = None

        if isinstance(code, Code):
            code_fields = code._head + (code.hash_bytes,)
        elif isinstance(code, str):
            code_fields = read_header(decode_base32(code))
        elif isinstance(code, tuple):
            code_fields = code
        elif isinstance(code, bytes):
            code_fields = read_header(code)
        else:
            raise ValueError(f"Code must be str, bytes, tuple or Code not {type(code)}")

        self._head = code_fields[:-1]
        body = bitarray()
        body.frombytes(code_fields[-1])
        self._body = frozenbitarray(body)

    def __str__(self):
        return self.code

    def __repr__(self):
        return f'Code("{self.code}")'

    def __bytes__(self):
        return self.bytes

    def __iter__(self):
        for f in self._head:
            yield f
        yield self.hash_bytes

    @property
    def code(self) -> str:
        """Standard base32 representation of code."""
        return encode_base32(self.bytes)

    @property
    def bytes(self) -> bytes:
        """Raw bytes of code (including header)."""
        return self.header_bytes + self._body.tobytes()

    @property
    def hex(self) -> str:
        """Hex representation of code (including header)."""
        return self.bytes.hex()

    @property
    def uint(self) -> int:
        """Integer representation of code (including header)"""
        return int.from_bytes(self.bytes, "big", signed=False)

    @property
    def type_id(self) -> str:
        """A unique composite type-id (use as name to index codes seperately)."""
        return (
            f"{self.maintype.name}-"
            f"{self.subtype.name}-"
            f"{self.version.name}-"
            f"{self.length}"
        )

    @property
    def explain(self) -> str:
        """Human readble representation of code header."""
        return f"{self.type_id}-{self.hash_hex}"

    @property
    def hash_bytes(self) -> bytes:
        """Byte representation of code (without header)"""
        return self._body.tobytes()

    @property
    def hash_hex(self) -> str:
        """Hex string representation of code (without header)."""
        return ba2hex(self._body)

    @property
    def hash_bits(self) -> str:
        """String of 0,1 representing the bits of the code (without header)."""
        return self._body.to01()

    @property
    def hash_ints(self) -> List[int]:
        """List of 0,1 integers representing the bits of the code (without header)."""
        return self._body.tolist(True)

    @property
    def hash_uint(self) -> int:
        """Unsinged integer representation of the code (without header)."""
        return ba2int(self._body, signed=False)

    @property
    def hash_ba(self) -> frozenbitarray:
        """Bitarray object of the code (without header)."""
        return self._body

    @property
    def header_bytes(self) -> bytes:
        """Byte representation of header prefix of the code"""
        return write_header(*self._head)

    @property
    def maintype(self) -> MT:
        """Enum maintype of code."""
        return MT(self._head[0])

    @property
    def subtype(self) -> Union[ST, ST_CC]:
        """Enum subtype of code."""
        if self.maintype in (MT.CONTENT, MT.ISCC):
            return ST_CC(self._head[1])
        return ST(self._head[1])

    @property
    def version(self) -> VS:
        """Enum version of code."""
        return VS(self._head[2])

    @property
    def length(self) -> int:
        """Length of code hash in number of bits (without header)."""
        return LN(self._head[3])

    @classmethod
    def rnd(cls, mt=None, bits=64, data=None):
        """Returns a syntactically correct random code (no SID support yet)"""

        # MainType
        mt = choice(list(MT)) if mt is None else mt

        # SubType
        if mt == MT.CONTENT:
            st = choice(list(ST_CC))
        else:
            st = choice(list(ST))

        # Version
        vs = VS.V0

        # Length
        ln = bits or choice(list(LN)).value

        # Body
        data = urandom(ln // 8) if data is None else data

        return cls((mt, st, vs, ln, data))

    def __xor__(self, other) -> int:
        """Use XOR operator for hamming distance calculation."""
        return count_xor(self._body, other._body)

    def __eq__(self, other):
        # type: (Code) -> bool
        return self.code == other.code

    def __hash__(self):
        return self.uint


def compose(codes):
    # type: (Iterable[AnyISCC]) -> Code
    """Combine/compress componets to a code of type ISCC (canonical 4-component form).

    :param Iterable[AnyISCC] codes: A sequence of Meta, Content, Data, Instance codes.
    """
    codes = sorted([Code(c) for c in codes], key=attrgetter("maintype"))
    assert len(codes) == 4, "ISCC composition requires 4 codes"
    assert len(set(c.version for c in codes)) == 1, "Codes must have same version"

    types = tuple(c.maintype for c in codes)
    assert MT.ISCC not in types, "Cannot compose canonical ISCC code"

    assert types == (
        MT.META,
        MT.CONTENT,
        MT.DATA,
        MT.INSTANCE,
    ), "Codes must be META, CONTENT, DATA, INSTANCE"

    for code in codes:
        assert code.length >= LN.L64, "ISCC requires min 64-bit codes"
        chash = b""
        for c in codes:
            chash += c.hash_bytes[:8]
        return Code((MT.ISCC, codes[1].subtype, codes[1].version, LN.L256, chash))


def decompose(iscc_code):
    # type: (AnyISCC) -> List[Code]
    """Decompose a canonical ISCC into a list of singular componet codes."""

    if isinstance(iscc_code, str):
        iscc_code = clean(iscc_code)

    code_obj = Code(iscc_code)
    raw = code_obj.bytes
    # Read sequence of components from raw bytes:
    components = []
    while raw:
        t, s, v, l, d = read_header(raw)
        if t == MT.ISCC:
            assert l == 256, "Code of type ISCC must be 256-bits"
            mco = Code((MT.META, ST.NONE, VS.V0, LN.L64, d[:8]))
            cco = Code((MT.CONTENT, s, VS.V0, LN.L64, d[8:16]))
            dco = Code((MT.DATA, ST.NONE, VS.V0, LN.L64, d[16:24]))
            ico = Code((MT.INSTANCE, ST.NONE, VS.V0, LN.L64, d[24:32]))
            components.extend([mco, cco, dco, ico])
            raw = d[32:]
        else:
            nbytes = l // 8
            body, raw = d[:nbytes], d[nbytes:]
            components.append(Code((t, s, v, l, body)))

    return components



__version__ = "0.1.2"
from iscc_core.code_meta import gen_meta_code, gen_meta_code_v0, soft_hash_meta_v0
from iscc_core.code_content_text import (
    gen_text_code,
    gen_text_code_v0,
    soft_hash_text_v0,
)
from iscc_core.code_content_image import (
    gen_image_code,
    gen_image_code_v0,
    soft_hash_image_v0,
)
from iscc_core.code_content_audio import (
    gen_audio_code,
    gen_audio_code_v0,
    soft_hash_audio_v0,
)
from iscc_core.code_content_video import (
    gen_video_code,
    gen_video_code_v0,
    soft_hash_video_v0,
)
from iscc_core.code_data import gen_data_code, gen_data_code_v0, soft_hash_data_v0
from iscc_core.code_instance import (
    gen_instance_code,
    gen_instance_code_v0,
    hash_instance_v0,
)
from iscc_core.codec import Code, compose, decompose



# -*- coding: utf-8 -*-
"""
A similarity preserving hash for text content (soft hash).
The **ISCC Content-Code Text** is generated from plain-text that has been extracted
from different media assets.
"""
import unicodedata
from typing import Union
import xxhash
from iscc_core.minhash import minhash_256
from iscc_core.models import ContentCodeText
from iscc_core.options import opts
from iscc_core.utils import sliding_window
from iscc_core import codec


Text = Union[str, bytes]


def gen_text_code(text, bits=opts.text_bits):
    # type: (Text, int) -> ContentCodeText
    """Create an ISCC Content-Code-Text with the latest standard algorithm.

    !!! note
        If `text` input includes markup (like HTML tags) it must be removed beforehand.

    :param Text text: Plain text for Text-Code creation.
    :param int bits: Bit-length (multiple of 32) for ISCC Code Hash (default 64).
    :return: TextCode with properties: code, characters
    :rtype: ContentCodeText
    """
    return gen_text_code_v0(text, bits)


def gen_text_code_v0(text, bits=opts.text_bits):
    # type: (Text, int) -> ContentCodeText
    """Create ISCC Content-Code-Text with algorithm v0

    - Normalize and lowercase text.
    - Create a Simhash from the tokens generated by a 13-character sliding window.
    - Encode resulting byte hash with ISCC codec.

    :param Text text: Normalized text for Text-Code creation.
    :param int bits: Bit-length of ISCC Code Hash (default 64).
    :return: TextCode with properties: code, characters
    :rtype: ContentCodeText
    """
    text_norm = normalize_text(text)
    characters = len(text_norm)
    text_lower = text_norm.lower()
    digest = soft_hash_text_v0(text_lower)
    text_code = codec.encode_component(
        mtype=codec.MT.CONTENT,
        stype=codec.ST_CC.TEXT,
        version=codec.VS.V0,
        length=bits,
        digest=digest,
    )
    return ContentCodeText(code=text_code, characters=characters)


def normalize_text(text):
    # type: (Text) -> str
    """Unicode normalization and character filtering.

    - Decode to Unicode.
    - Remove leading/trailing whitespace.
    - Decompose with NFD normalization.
    - Filter special characters and whitespace.
    - Remove duplicate whitespace.
    - Recombine with NFKC normalization.

    :param Text text: Plain text to be normalized.
    :return: Normalized plain text.
    :rtype: str
    """

    # 1. Convert bytes to str
    if isinstance(text, bytes):
        text = text.decode("utf-8")

    # 2. Remove leading/trailing whitespace
    text_stripped = text.strip()

    # 3. Decompose with NFD
    text_decomposed = unicodedata.normalize("NFD", text_stripped)

    # 4. Filter
    chars = []
    for c in text_decomposed:
        cat = unicodedata.category(c)
        if cat not in opts.text_unicode_filter:
            chars.append(c)
        elif c in opts.text_whitespace:
            chars.append(c)
    text_filtered = "".join(chars)

    # 5. Collapse consecutive whitespace
    wsproc_text = " ".join(text_filtered.split())

    # 6. Recombine
    recombined = unicodedata.normalize("NFKC", wsproc_text)

    return recombined


def soft_hash_text_v0(text):
    # type: (str) -> bytes
    """
    Create a 256-bit similarity preserving hash for text input with v0 algorithm.
    Text should be stripped of markup, normalized and lowercased before hash creation.

    :param str text: Plain text to be hashed.
    :return: 256-bit similarity preserving byte hash
    :rtype: bytes
    """
    ngrams = sliding_window(text, opts.text_ngram_size)
    features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
    hash_digest = minhash_256(features)
    return hash_digest



# -*- coding: utf-8 -*-
from typing import Generator, Sequence


def sliding_window(seq, width):
    # type: (Sequence, int) -> Generator
    """
    Generate a sequence of equal "width" slices each advancing by one elemnt.
    All types that have a length and can be sliced are supported (list, tuple, str ...).
    The result type matches the type of the input sequence.
    Fragment slices smaller than the width at the end of the sequence are not produced.
    If "witdh" is smaller than the input sequence than one element will be returned that
    is shorter than the requested width.

    :param Sequence seq: Sequence of values to slide over
    :param int width: Width of sliding window in number of items
    :returns: A generator of window sized items
    :rtype: Generator
    """
    assert width >= 2, "Sliding window width must be 2 or bigger."
    idx = range(max(len(seq) - width + 1, 1))
    return (seq[i : i + width] for i in idx)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iscc_core']

package_data = \
{'': ['*']}

install_requires = \
['Pillow',
 'bitarray-hardbyte',
 'blake3',
 'more-itertools',
 'pydantic>=1.8.2,<2.0.0',
 'xxhash']

extras_require = \
{'turbo': ['cython', 'pybase64']}

setup_kwargs = {
    'name': 'iscc-core',
    'version': '0.1.2',
    'description': 'ISCC - Core Algorithms',
    'long_description': '# iscc-core - ISCC Core Algorithms\n\n[![Build](https://github.com/iscc/iscc-core/actions/workflows/tests.yml/badge.svg)](https://github.com/iscc/iscc-core/actions/workflows/tests.yml)\n\n> **iscc-core** is a Python library that implements the core algorithms of [ISCC v1.1](https://iscc.codes)\n(International Standard Content Code)\n\n## What is ISCC\n\nThe **ISCC** (*International Standard Content Code*) is an identifier for digital media\nassets.\n\nAn **ISCC** is derived algorithmically from the digital content itself, just like\ncryptographic hashes. However, instead of using a single cryptographic hash function to\nidentify data only, the ISCC uses a variety of algorithms to create a composite\nidentifier that exhibits similarity-preserving properties (soft hash).\n\nThe component-based structure of the ISCC identifies content at multiple levels of\nabstraction. Each component is self-describing, modular and can be used separately or\nin conjunction with others to aid in various content identification tasks.\n\nThe algorithmic design supports scenarios that require content deduplication, database\nsynchronisation and indexing, integrity verification, timestamping, versioning, data\nprovenance, similarity clustering, anomaly detection, usage tracking, allocation of\nroyalties, fact-checking and general digital asset management use-cases.\n\n## What is `iscc-core`\n\n`iscc-core` is the python based library of the core algorithms to create standard\ncompliant **ISCC** codes. It also serves as a reference for porting ISCC to other\nprogramming languages.\n\n## ISCC Architecture\n\n![ISCC Architecure](./docs/images/iscc-architecture.png)\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install `iscc-core`.\n\n```bash\npip install iscc-core\n```\n\n## Quick Start\n\n```python\nfrom iscc_core import (\n    gen_meta_code,\n    gen_image_code,\n    gen_data_code,\n    gen_instance_code,\n    compose,\n)\n\nimage_path = "../docs/images/iscc-architecture.png"\n\nmeta_code = gen_meta_code(\n    title="ISCC Architecure",\n    extra="A schematic overview of the ISCC"\n)\n\nprint("Meta-Code:\\t\\t", meta_code.code)\nprint("Structure:\\t\\t", meta_code.code_obj.explain, end="\\n\\n")\n\nwith open(image_path, "rb") as stream:\n\n    image_code = gen_image_code(stream)\n    print("Image-Code:\\t\\t", image_code.code)\n    print("Structure:\\t\\t", image_code.code_obj.explain, end="\\n\\n")\n\n    stream.seek(0)\n    data_code = gen_data_code(stream)\n    print("Data-Code:\\t\\t", data_code.code)\n    print("Structure:\\t\\t", data_code.code_obj.explain, end="\\n\\n")\n\n    stream.seek(0)\n    instance_code = gen_instance_code(stream)\n    print("Instance-Code:\\t", instance_code.code)\n    print("Structure:\\t\\t", instance_code.code_obj.explain, end="\\n\\n")\n\niscc_code = compose((meta_code.code, image_code.code, data_code.code, instance_code.code))\nprint("Canonical ISCC:\\t ISCC:{}".format(iscc_code.code))\nprint("Structure:\\t\\t", iscc_code.explain)\n```\n\nThe output of this example is as follows:\n\n```\nMeta-Code:       AAA5H3V6SZHWDUKX\nStructure:       META-TEXT-V0-64-d3eebe964f61d157\n\nImage-Code:      EEA34YXAFUJWOZ5Q\nStructure:       CONTENT-IMAGE-V0-64-be62e02d136767b0\n\nData-Code:       GAA6JYHWISAVU77Z\nStructure:       DATA-NONE-V0-64-e4e0f644815a7ff9\n\nInstance-Code:   IAAUULVLVWQLXSEM\nStructure:       INSTANCE-NONE-V0-64-4a2eabada0bbc88c\n\nCanonical ISCC:  ISCC:KED5H3V6SZHWDUKXXZROALITM5T3BZHA6ZCICWT77FFC5K5NUC54RDA\nStructure:       ISCC-IMAGE-V0-256-d3eebe964f61d157be62e02d136767b0e4e0f644815a7ff94a2eabada0bbc88c\n```\n\n## Documentation\n\nhttps://iscc-core.iscc.codes\n\n## Project Status\n\nISCC is in the process of being standardized within\n[ISO/TC 46/SC 9](https://www.iso.org/standard/77899.html).\n\n## Maintainers\n[@titusz](https://github.com/titusz)\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to discuss\nwhat you would like to change. Please make sure to update tests as appropriate.\n\nYou may also want join our developer chat on Telegram at <https://t.me/iscc_dev>.\n\n\n## Changelog\n\n### [0.1.2] - 2021-11-14\n- Export more functions to toplevel\n- Return schema driven objects from ISCC code generators.\n\n### [0.1.1] - 2021-11-14\n- Fix packaging problems\n\n### [0.1.0] - 2021-11-13\n- Initial release\n\n\n',
    'author': 'Titusz',
    'author_email': 'tp@py7.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
